﻿B4J=true
Group=Default Group
ModulesStructureVersion=1
Type=Class
Version=8.1
@EndOfDesignText@
Sub Class_Globals
	Private Buckets(200) As List
	Type HTKeyValue (Key As String, Value As Object)
End Sub

Public Sub Initialize 
	For i = 0 To Buckets.Length - 1
		Buckets(i).Initialize
	Next
End Sub

Public Sub Clear
	For i = 0 To Buckets.Length - 1
		Buckets(i).Clear
	Next
End Sub

'Puts the key / value pair in the table. Replaces any existing value with the same key.
Public Sub Put (Key As String, Value As Object)
	Dim b As Int = FindBucket(Key)
	Dim i As Int = FindItemIndexInBucket(b, Key)
	If i >-1 Then
		'existing item exists. Replace the value.
		Dim kv As HTKeyValue = Buckets(b).Get(i)
		kv.Value = Value
	Else
		'add new item to the bucket list.
		Buckets(b).Add(CreateHTKeyValue(Key, Value))
	End If
End Sub

'Gets the value associated with the given key. Returns Null if not exists.
Public Sub Get (Key As String) As Object
	Dim b As Int = FindBucket(Key)
	Dim i As Int = FindItemIndexInBucket(b, Key)
	If i = -1 Then Return Null
	Dim kv As HTKeyValue = Buckets(b).Get(i)
	Return kv.Value
End Sub

'Checks whether the given key exists in the table.
Public Sub ContainsKey (Key As String) As Boolean
	Dim b As Int = FindBucket(Key)
	Dim i As Int = FindItemIndexInBucket(b, Key)
	Return i > -1
End Sub

'Returns all keys. Note that the order is not preserved.
Public Sub GetKeys As List
	Dim res As List
	res.Initialize
	For Each bucket As List In Buckets
		For Each kv As HTKeyValue In bucket
			res.Add(kv.Key)
		Next
	Next
	Return res
End Sub

'Removes the item with the given key, if such exists.
Public Sub Remove (Key As String)
	Dim b As Int = FindBucket(Key)
	Dim i As Int = FindItemIndexInBucket(b, Key)
	If i > -1 Then Buckets(b).RemoveAt(i)
End Sub

Private Sub FindItemIndexInBucket(BucketNumber As Int, Key As String) As Int
	For i = 0 To Buckets(BucketNumber).Size - 1
		Dim kv As HTKeyValue = Buckets(BucketNumber).Get(i)
		If kv.Key = Key Then Return i
	Next
	Return -1
End Sub

Private Sub FindBucket (key As String) As Int
	Return Abs(CalcHash (key)) Mod Buckets.Length
End Sub

Private Sub CalcHash (key As String) As Int
	'djb2 hash algorithm
	Dim h As Int = 5381
	For i = 0 To key.Length - 1
		h = Bit.ShiftLeft(h, 5) + h + Asc(key.CharAt(i))
	Next
'	Log(key &": " & h)
	Return h
End Sub


Private Sub CreateHTKeyValue (Key As String, Value As Object) As HTKeyValue
	Dim t1 As HTKeyValue
	t1.Initialize
	t1.Key = Key
	t1.Value = Value
	Return t1
End Sub

Public Sub LogBuckets
	Dim sb As StringBuilder
	sb.Initialize
	For i = 0 To Buckets.Length - 1
		If Buckets(i).Size > 0 Then
			sb.Append(i).Append(": ")
			For Each kv As HTKeyValue In Buckets(i)
				sb.Append(kv.Key).Append(", ")
			Next
			'remove the last comma.
			sb.Remove(sb.Length - 2, sb.Length)
			sb.Append(CRLF)
		End If
	Next
	Log(sb.ToString)
End Sub